from openai import BadRequestError, APITimeoutError, RateLimitError
import tiktoken
from openai import AzureOpenAI, OpenAI
import re
import time
from tqdm import tqdm
from pathlib import Path
from loguru import logger
from typing import Optional
from .support_api import SupportAPI
from .cache_manager import CacheManager


class WrapOpenAI:
    def __init__(
        self,
        api_type: str = "azure_openai",
        api_base: str = "",
        api_version: str = "2024-02-01",
        api_key: str = "",
        modelname: Optional[str] = None,
        temperature: Optional[float] = None,
        top_p: Optional[float] = None,
        max_tokens: Optional[int] = None,
        timeout: int = 600,
        enable_cache: bool = False,
        cache_dir: Optional[str] = None,
        cache_prefix: Optional[str] = None,
        cache_interval: int = 0
    ):
        try:
            self.api_type = SupportAPI(api_type)
        except KeyError:
            raise ValueError(f"{api_type} is not a valid name for {SupportAPI.__name__}")

        self.api_base = api_base
        self.api_version = api_version
        self.api_key = api_key
        self.modelname = modelname
        self.temperature = temperature
        self.top_p = top_p
        self.max_tokens = max_tokens
        self.timeout = timeout
        self.enable_cache = enable_cache
        self.cache_dir = cache_dir
        self.cache_prefix = cache_prefix
        self.cache_interval = cache_interval

        if self.api_type == SupportAPI.AzureOpenAI:
            self._client = AzureOpenAI(
                api_version=self.api_version,
                azure_endpoint=self.api_base,
                api_key=self.api_key,
            )
        elif self.api_type in [SupportAPI.OpenAI, SupportAPI.OpenAILike]:
            #   "openai" or "openai_like"
            self._client = OpenAI(
                base_url=self.api_base,
                api_key=self.api_key,
            )

        self._log_format_parameters()
        self._init_cache()
        self.encoding_init_completed = False

    def _log_format_parameters(self):
        """Helper method to format parameters with masked api_key."""
        masked_api_key = '*' * len(self.api_key) if self.api_key else ''
        params = {
            "api_type": self.api_type,
            "api_base": self.api_base,
            "api_version": self.api_version,
            "api_key": masked_api_key,
            "modelname": self.modelname,
            "temperature": self.temperature,
            "top_p": self.top_p,
            "max_tokens": self.max_tokens,
            "timeout": self.timeout,
            "enable_cache": self.enable_cache,
            "cache_dir": self.cache_dir,
            "cache_prefix": self.cache_prefix
        }

        # Formatting parameters for printing
        formatted_params = "\n".join(f"{key}: {value}" for key, value in params.items())
        logger.info(f"{formatted_params}")

    def _init_encoding(self):
        if self.api_type in [SupportAPI.AzureOpenAI, SupportAPI.OpenAI]:
            if self.modelname.startswith('gpt-4o'):
                self.enconding_name = "o200k_base"
                self.encoding = tiktoken.get_encoding(self.enconding_name)
            elif self.modelname.startswith('gpt-4') or self.modelname.startswith('gpt-3.5'):
                self.enconding_name = "cl100k_base"
                self.encoding = tiktoken.get_encoding(self.enconding_name)
            else:
                raise ValueError(f"modelname == {self.modelname}")
        else:
            raise ValueError("When `api_type` is not equal to azure_openai or openai, it can't work.")

        self.encoding_init_completed = True

    def _init_cache(self):
        if self.enable_cache and self.cache_dir:
            self.cache_dir = Path(self.cache_dir)
            self.cache_dir.mkdir(exist_ok=True, parents=True)
            if self.cache_prefix:
                cache_prefix = self.cache_prefix
            else:
                if self.modelname:
                    cache_prefix = "-".join(self.modelname.split("-")[:2])
                else:
                    cache_prefix = "default"

            cache_path = self.cache_dir / f"{cache_prefix}.cache"
            self.gpt_cache = CacheManager(cache_path, backup_interval=self.cache_interval)

    def infer(self, prompt):
        response = self._client.chat.completions.create(
            model=self.modelname,
            max_tokens=self.max_tokens,
            temperature=self.temperature,
            top_p=self.top_p,
            timeout=self.timeout,
            messages=[
                {"role": "system", "content": prompt},
            ]
        )

        if response is not None \
           and hasattr(response, "choices") \
           and len(response.choices) > 0 \
           and hasattr(response.choices[0], "message") \
           and hasattr(response.choices[0].message, "content"):

            response = response.choices[0].message.content

        return response

    def repeat_complete(self, complete_fn, prompt, sleep_eps=60, max_retry=3, every_step_sleep=0):
        resp = None

        for _ in range(max_retry):
            try:
                resp = complete_fn(prompt)
                time.sleep(every_step_sleep)
                break
            except BadRequestError as e:
                logger.warning(f"BadRequestError: {repr(e)}")
                logger.warning(f"{prompt}")
                break
            except APITimeoutError as e:
                logger.warning(f"APITimeoutError: {repr(e)}，delay {sleep_eps}s and retry.")
                time.sleep(sleep_eps)
            except RateLimitError as e:
                err_msg = repr(e)
                reg = r"Please retry after (\d+) seconds"
                nums = re.findall(reg, err_msg)
                if len(nums) != 1:
                    time.sleep(sleep_eps)
                    continue
                num = nums[0]
                if not num.isdigit():
                    time.sleep(sleep_eps)
                server_eps = int(num)
                logger.warning(f"{repr(e)}.\nExtraction time delayed, retrying after {server_eps}s.")
                time.sleep(server_eps)

            except Exception as e:
                logger.error(f"{repr(e)}, retrying after a delay of {sleep_eps}s.")
                time.sleep(sleep_eps)

        return resp

    def num_tokens_from_string(self, string: str) -> int:
        if not self.encoding_init_completed:
            self._init_encoding()

        num_tokens = len(self.encoding.encode(string))
        return num_tokens

    def chatcomplete(self, prompts, verbose=False):
        responses = [None] * len(prompts)

        if verbose:
            prompts = tqdm(prompts)

        if self.enable_cache:
            for idx, prompt in enumerate(prompts):
                response = self.gpt_cache.get_item(prompt)
                if response is not None:
                    responses[idx] = response

        for idx, (prompt, cached_response) in enumerate(zip(prompts, responses)):
            logger.debug(f"## input\n{prompt}")
            if cached_response is not None:
                logger.debug(f"## response(cached)\n{cached_response}")
            else:
                response = self.repeat_complete(self.infer, prompt, sleep_eps=10, max_retry=300, every_step_sleep=2)

                if response is not None:
                    logger.debug(f"## response(new)\n{response}")
                    responses[idx] = response
                    if self.enable_cache:
                        self.gpt_cache.set_item(key=prompt, value=response)
                else:
                    logger.debug("## response(new)\nNone")

        return responses



from pathlib import Path
import sqlite3
from loguru import logger
import time
import shutil


class CacheManager:
    def __init__(self, cache_path, backup_interval=0):
        logger.info(f"Cache is in: {cache_path}")
        self.cache_path = Path(cache_path)
        is_first_run = not self.cache_path.exists()
        self.conn = sqlite3.connect(cache_path)
        self.conn.execute('PRAGMA journal_mode=WAL')
        self.cursor = self.conn.cursor()
        if is_first_run:
            self.create_table()
        self.last_backup_time = 0
        self.backup_interval = backup_interval  # 1 hour

    def create_table(self):
        self.cursor.execute('''
            CREATE TABLE IF NOT EXISTS kv_cache (
                key TEXT PRIMARY KEY,
                value TEXT NOT NULL
            )
        ''')
        self.conn.commit()

    def count(self):
        """
        Query the current number of records in the kv_cache table.
        """
        self.cursor.execute('SELECT COUNT(*) FROM kv_cache')
        count = self.cursor.fetchone()[0]
        logger.info(f"Now records count: {count}")

    def set_item(self, key, value):
        """
        Insert or update a key-value pair in the kv_cache table.

        If the provided value is None or an empty string, a warning is logged,
        and the operation is aborted. Otherwise, the method attempts to insert
        the key-value pair into the kv_cache table. If the key already exists,
        the value is updated with the new value.

        After a successful insertion or update, the current record count is logged.
        In case of an SQLite error during the operation, an error message is logged,
        and the transaction is rolled back to maintain data integrity.
        Args:
            key (str): The key to insert or update in the cache.
            value (str): The value to associate with the key.

        Raises:
            sqlite3.Error: If there is a database error during the insertion or update.
        """
        if value is None or (isinstance(value, str) and value.strip() == ""):
            logger.warning(f"Value is None or empty for key: {key}")
            return
        try:
            self.cursor.execute('''
                INSERT INTO kv_cache (key, value) VALUES (?, ?)
                ON CONFLICT(key) DO UPDATE SET value=excluded.value
            ''', (key, value))
            self.conn.commit()
            self.count()
            self.backup_cache()
        except sqlite3.Error as e:
            logger.error(f"Error inserting key-value: {e}")
            self.conn.rollback()

    def get_item(self, key):
        self.cursor.execute('''
            SELECT value FROM kv_cache WHERE key=?
        ''', (key,))
        result = self.cursor.fetchone()
        return result[0] if result else None

    def check_integrity(self):
        try:
            self.cursor.execute("PRAGMA integrity_check")
            result = self.cursor.fetchone()
            return result[0] == "ok"
        except sqlite3.Error as e:
            logger.error(f"Error checking database integrity: {e}")
            return False

    def backup_cache(self):
        current_time = time.time()
        if current_time - self.last_backup_time < self.backup_interval:
            return

        if self.check_integrity():
            backup_path = self.cache_path.with_suffix('.bak')
            try:
                shutil.copy2(self.cache_path, backup_path)
                logger.info(f"Cache backed up to {backup_path}")
                self.last_backup_time = current_time
            except IOError as e:
                logger.error(f"Error backing up cache: {e}")
        else:
            logger.error("Cache integrity check failed. Backup not performed.")

    def delete(self, key):
        self.cursor.execute('''
            DELETE FROM kv_cache WHERE key=?
        ''', (key,))
        self.conn.commit()

    def update(self, key, value):
        self.set(key, value)  # Reuse set method with conflict update logic

    def close(self):
        self.backup_cache()
        self.conn.close()

    def __del__(self):
        self.close()



from typing import List, Dict
from . import io_util
from .wrap_openai import WrapOpenAI
from loguru import logger
from pathlib import Path
import os
from jinja2 import Template


class LLMQuiver:
    def __init__(self) -> None:
        self._initialize_by_env()
        self.prompt_template = None

    def _initialize_by_env(self):
        LLMQUIVER_CONFIG = os.environ.get("LLMQUIVER_CONFIG", None)
        if LLMQUIVER_CONFIG is None:
            raise ValueError(f"LLMQUIVER_CONFIG is invalid: {LLMQUIVER_CONFIG}.")

        LLMQUIVER_CONFIG = Path(LLMQUIVER_CONFIG)
        if LLMQUIVER_CONFIG.exists():
            logger.info(f"LLMQUIVER_CONFIG: {LLMQUIVER_CONFIG}")
            config = io_util.read_toml(LLMQUIVER_CONFIG)
        else:
            raise ValueError(f"LLMQUIVER_CONFIG: {LLMQUIVER_CONFIG} is not found.")

        keys = ["API_TYPE", "MODEL_NAME", "API_BASE", "API_KEY", "API_VERSION"]
        params = {}
        for key in keys:
            env_value = os.environ.get(key, None)
            config_value = config.get(key, None)
            if env_value and len(env_value) > 0:
                params[key] = env_value
            elif config_value and len(config_value) > 0:
                params[key] = config_value
            else:
                raise ValueError(f"Missing {key} in both the configuration file({LLMQUIVER_CONFIG}) and environment.")

        self.gen = WrapOpenAI(
            api_type=params["API_TYPE"],
            api_base=params["API_BASE"],
            api_version=params["API_VERSION"],
            api_key=params["API_KEY"],
            modelname=params["MODEL_NAME"],
            temperature=config.get("temperature"),
            top_p=config.get("top_p", None),
            max_tokens=config.get("max_tokens"),
            enable_cache=config.get("enable_cache"),
            cache_dir=config.get("cache_dir"),
            cache_prefix=config.get("cache_prefix", params["MODEL_NAME"]),
            cache_interval=config.get("cache_interval", 0),
        )

    def get_num_tokens_from_string_fn(self):
        if hasattr(self.gen, "num_tokens_from_string"):
            return self.gen.num_tokens_from_string
        else:
            raise NotImplementedError("num_tokens_from_string is not implemented.")

    def prerpare_prompt(self, prompt_values):
        if not isinstance(prompt_values, list):
            raise TypeError("prompt_values must be a list.")

        if len(prompt_values) == 0:
            return []

        if self.__class__.__name__ == "LLMQuiver":
            if not isinstance(prompt_values[0], str):
                raise TypeError("LLMQuiver's prompt_values must be a list of str.")
        elif self.__class__.__name__ == "Jinja2LLMQuiver" or self.__class__.__name__ == "TomlLLMQuiver":
            if not isinstance(prompt_values[0], dict):
                raise TypeError("Jinja2LLMQuiver's or TomlLLMQuiver's prompt_values must be a list of dict.")

        return [self.render_prompt(p) for p in prompt_values]

    def render_prompt(self, prompt_value):
        return prompt_value

    def generate_by_toml_template(self, prompts):
        candidates = []
        cand_idx = []
        responses = [None] * len(prompts)
        for i, prompt in enumerate(prompts):
            cand_idx.append(i)
            candidates.append(prompt)

        model_outs = self.gen.chatcomplete(prompts=prompts, verbose=True)

        for i, out in zip(cand_idx, model_outs):
            responses[i] = out

        return responses

    def generate(
        self, prompt_values: List[Dict]
    ):
        prompts = self.prerpare_prompt(prompt_values)
        return self.gen.chatcomplete(prompts=prompts, verbose=True)


class Jinja2LLMQuiver(LLMQuiver):
    def __init__(
        self,
        jinja2_template_file=None
    ):
        super().__init__()

        if jinja2_template_file is None:
            raise ValueError("jinja2_template_file is required for jinja2 template type.")

        jinja2_template_file = Path(jinja2_template_file)
        logger.info(f"jinja2_template_file: {jinja2_template_file}")

        if jinja2_template_file.exists():
            template_content = io_util.read_text(jinja2_template_file)
            self.prompt_template = Template(template_content)
        else:
            raise ValueError(f"jinja2_template_file: {jinja2_template_file} is not found.")

    def render_prompt(self, prompt_value):
        return self.prompt_template.render(prompt_value)


class TomlLLMQuiver(LLMQuiver):
    def __init__(
        self,
        toml_template_file=None,
        toml_prompt_name=None
    ):

        super().__init__()

        if toml_template_file is None or toml_prompt_name is None:
            raise ValueError("toml_template_file is required for toml template type.")
        if not isinstance(toml_prompt_name, str) or len(toml_prompt_name) == 0:
            raise ValueError("toml_prompt_name is required for toml template type.")

        toml_template_file = Path(toml_template_file)
        logger.info(f"basic_toml_file: {toml_template_file}, basic_prompt_name: {toml_prompt_name}")

        if not toml_template_file.exists():
            raise ValueError(f"toml_template_file: {toml_template_file} is not found.")

        prompt_templ_map = io_util.read_toml(toml_template_file)
        if toml_prompt_name not in prompt_templ_map:
            raise ValueError(f"Can't find a template named '{toml_prompt_name}'.")

        self.prompt_template = prompt_templ_map[toml_prompt_name]

    def render_prompt(self, prompt_value):
        return self.prompt_template.format_map(prompt_value)



import yaml
import toml
import json
import os


def write_text(text, dst):
    with open(dst, 'w', encoding='utf-8') as file:
        file.write(text)


def read_text(src):
    with open(src, 'r') as file:
        return file.read()


def read_jsonl(src):
    with open(src, 'r') as json_file:
        res = []
        json_list = list(json_file)
        for json_str in json_list:
            res.append(json.loads(json_str))
        return res


def write_jsonl(obj, dst):
    with open(dst, 'w') as json_file:
        res = []
        for json_str in obj:
            res.append(json.dumps(json_str, ensure_ascii=False))
        json_file.write("\n".join(res))


def write_json(obj, dst):
    with open(dst, "w", encoding="utf-8") as f:
        json.dump(obj, f, ensure_ascii=False, indent=2)


def read_json(src):
    with open(src, "r", encoding="utf-8") as f:
        obj = json.load(f)
    return obj


def read_yaml(src):
    with open(src, 'r', encoding='utf-8') as file:
        try:
            data = yaml.safe_load(file)
            return data
        except yaml.YAMLError as exc:
            print(exc)


def convert_yaml_to_obj(yaml_str):
    return yaml.safe_load(yaml_str)


def convert_obj_to_yaml(obj):
    return yaml.dump(obj, indent=2, allow_unicode=True, sort_keys=False)


def write_yaml(obj, dst):
    with open(dst, 'w', encoding='utf-8') as file:
        try:
            yaml.safe_dump(obj, file, allow_unicode=True, sort_keys=False)
        except yaml.YAMLError as exc:
            print(exc)


def read_toml(src):
    with open(src, 'r') as toml_file:
        data = toml.load(toml_file)
    return data


def write_toml(obj, dst):
    with open(obj, 'w') as toml_file:
        toml.dump(dst, toml_file)


def read_file_lst(src, endswith=None):
    filenames = os.listdir(src)
    if endswith:
        filenames = filter(lambda x: x.endswith(endswith), filenames)
    return list(map(lambda x: os.path.join(src, x), filenames))
